/******************************************************************************
*                                                                             *
* License Agreement                                                           *
*                                                                             *
* Copyright (c) 2014 Altera Corporation, San Jose, California, USA.           *
* All rights reserved.                                                        *
*                                                                             *
* Permission is hereby granted, free of charge, to any person obtaining a     *
* copy of this software and associated documentation files (the "Software"),  *
* to deal in the Software without restriction, including without limitation   *
* the rights to use, copy, modify, merge, publish, distribute, sublicense,    *
* and/or sell copies of the Software, and to permit persons to whom the       *
* Software is furnished to do so, subject to the following conditions:        *
*                                                                             *
* The above copyright notice and this permission notice shall be included in  *
* all copies or substantial portions of the Software.                         *
*                                                                             *
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  *
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    *
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE *
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER      *
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     *
* FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         *
* DEALINGS IN THE SOFTWARE.                                                   *
*                                                                             *
* This agreement shall be governed in all respects by the laws of the State   *
* of California and by the laws of the United States of America.              *
* Altera does not recommend, suggest or require that this reference design    *
* file be used in conjunction or combination with any other product.          *
******************************************************************************/

/**********************************************************************
 *
 * Filename:    msgdma_software_example.c
 *
 * Version 1.0 (revision release)
 *
 * Description:
 *
 * This software uses the msgdma to move contents from one
 * location in memory to another in the same memory.
 * The software performs the following steps:
 *
 * 1)  Generates a memory buffer with 0x10h bytes aligned access and size 
 *     1024 bytes for both source and destination. The buffer lengths are 
 *     set to size of the buffer allocation.
 *
 * 2)  Source buffers are populated with an incrementing pattern of the 
 *     0, 1, 2, 3 ..., 255 , each number will occupy 4 bytes.
 *
 * 3)  Data cache is flushed to the memory before the extended descriptor is 
 *     constructed and send to the msgdma. The read stride field of the extended
 *     descriptor is set to "2" and the write stride field is set to "1".
 *
 * 4)  Start the data transfer with data width set to "32" bits, transfer type 
 *     set to "Aligned Access", with stride addressing, and extended feature 
 *     support enabled, max stride words set to "2", and max transfer length 
 *     set to "2KB" in qsys and calling the synchronous transfer API until all source 
 *     data are transferred into destination.
 *
 * 5)  Software waits for the transfer complete before continue processing. 
 *     However, time out 5 msec will be set to avoid trapping into looping forever. 
 *
 * 6)  Data in the destination buffer will be verified for the accuracy after 
 *     data transfer is completed and data cache is invalidated. The data pattern in 
 *     destination buffer are copied from source buffer with address incrementing 
 *     based on master data width. The data pattern in destination buffer could be 
 *     displayed as 0, 2, 4, 6, ....,254.
 *
 * 7)  Release the allocated buffer.
 **********************************************************************/

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include "altera_msgdma_descriptor_regs.h"
#include "altera_msgdma_csr_regs.h"
#include "altera_msgdma_response_regs.h"
#include "altera_msgdma.h"
#include "sys/alt_irq.h"
#include "alt_types.h"
#include "sys/alt_cache.h"
#include "sys/alt_errno.h"

#define HEAP_MEMORY_ALLOC_MAX 256

int main(void)
{

    alt_msgdma_extended_descriptor a_descriptor;
    alt_msgdma_extended_descriptor * a_descriptor_ptr = &a_descriptor;
    alt_msgdma_dev dev;
    alt_msgdma_dev *dev_ptr = &dev;
    alt_u32 *read_address;
    alt_u32 *write_address;
    alt_u32 j, i, data_read, control = 0;
    alt_u32 * ram_access_ptr;  /* using this to step through the buffers 4 byte at a time*/
    alt_u32 * ram_writeaccess_ptr;
    alt_32 status = 0;

    printf("Attempting to open MSGDMA...\n");
    dev_ptr = alt_msgdma_open(MSGDMA_M2M_STRIDE_CSR_NAME);

    if (NULL==dev_ptr)
    {
        printf("Cannot find the device with the name "
        	"<<MSGDMA_M2M_STRIDE_CSR_NAME>>\n");
        return 0;
    }



    /*allocate memory size of (HEAP_MEMORY_ALLOC_MAX*32) and 
      0x10h bytes aligned access*/
    read_address = memalign(16,HEAP_MEMORY_ALLOC_MAX * sizeof(alt_u32));
    write_address = memalign(16,HEAP_MEMORY_ALLOC_MAX * sizeof(alt_u32));


    if (NULL == read_address || NULL == write_address)
    {
        printf("Memory allocation failed.\n");
        return 0;
    }


    ram_access_ptr = read_address;

    /* populate the source data */
    for(j = 0; j < HEAP_MEMORY_ALLOC_MAX; j++)
    {
        /* writing a repeating pattern of 0, 1, .. ,0xFF, ..*/
        *ram_access_ptr++ = (j & 0xFFFFFFFF);
    }

        /* need to make sure all the data is written out to memory */
        alt_dcache_flush_all();

        /* constructing mm_to_mm extended descriptors */
      if(alt_msgdma_construct_extended_mm_to_mm_descriptor
	  (
	          dev_ptr,
              a_descriptor_ptr,
              read_address,
              write_address,
              j * sizeof(alt_u32),
              control,
              0x0000,
              0x0,
              0x0,
              0x02,
              0x01
          ) == -EINVAL)
        {
    	  printf(" invalid argument \n");
    	  goto exit;
        }

        /* start blocking transfer content from source address to destination address */
        status = alt_msgdma_extended_descriptor_sync_transfer(dev_ptr,a_descriptor_ptr);

        if(0 == status)
        {
            /* invalidate the data cache */
            alt_dcache_flush_no_writeback(write_address,j * sizeof(alt_u32));

            /* verify the data written by the msgdma at the destination addresses */
            ram_writeaccess_ptr = write_address;
            for(i = 0; i < j; i=i+2)
            {
                /* should be a pattern of 0, 1, .., 0x20, .. */
                data_read = *ram_writeaccess_ptr++;
                if ((i & 0xFFFFFFFF) != data_read)
                {
                    printf("base 0x%x offset_add 0x%x failed,expected %ld but read 0x%x\n"
                    	  ,write_address, ram_writeaccess_ptr,(i&0xFFFFFFFF), data_read);

                    goto exit;
                }
            }

        }
        else if(-EPERM == status)
        {
            printf("operation not permitted due to descriptor type conflict\n");
            goto exit;
        }
        else if (-ETIME == status)
        {
            printf("Time out and skipping the looping after 5 msec\n");
            goto exit;
        }
        else
        {
            printf("Errors flag = %ld\n", status);
            goto exit;
        }


    printf("Memory copy successful. Session will now close. \n");

    exit:
    free(read_address);
    free(write_address);

	printf("%c",(char)4);
    return 0;
}
